body {
  width: 60%;
  margin: 2em auto;
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
  line-height: 1.5;
  color: #24292e;
}

h1, h2 {
  border-bottom: 1px solid #eaecef;
  padding-bottom: 0.3em;
}

a {
  color: #0366d6;
  text-decoration: none;
}

code {
  padding: 0.2em 0.4em;
  margin: 0;
  background-color: rgba(27,31,35,0.05);
  border-radius: 3px;
  font-family: "SFMono-Regular", Consolas, "Liberation Mono", Menlo, Courier, monospace;
  font-size: 85%;
}

@media screen and (max-width: 1000px) {
  body { width: 75%; }
}

@media screen and (max-width: 700px) {
  body { width: 90%; }
}

